#!/usr/bin/python

import datetime
import os
import pycassa
import random
import unittest

from testtools import TestCase
from oopsrepository.testing.cassandra import TemporaryOOPSDB

from daisy import config
from daisy import schema
from oopsrepository import schema as oopsschema
from oopsrepository import config as oopsconfig

class TestCassandra(TestCase):
    def setUp(self):
        super(TestCassandra, self).setUp()

        # Set up daisy schema.
        os.environ['OOPS_HOST'] = config.cassandra_hosts[0]
        self.keyspace = self.useFixture(TemporaryOOPSDB()).keyspace
        os.environ['OOPS_KEYSPACE'] = self.keyspace
        config.cassandra_keyspace = self.keyspace
        schema.create()

        cfg = oopsconfig.get_config()
        cfg['username'] = config.cassandra_username
        cfg['password'] = config.cassandra_password
        oopsschema.create(cfg)

        creds = {'username': config.cassandra_username,
                 'password': config.cassandra_password}
        self.pool = pycassa.ConnectionPool(self.keyspace,
                                           config.cassandra_hosts,
                                           credentials=creds)
        from errors import cassandra
        self.cassandra = cassandra
        self.cassandra.pool = self.pool

class TestPackageCrashRate(TestCassandra):

    def test_increase(self):
        '''If a package has more crashes today than the average there is an
           increase.'''
        counters_cf = pycassa.ColumnFamily(self.pool, 'Counters')
        release = 'Ubuntu 12.10'
        package = 'whoopsie'
        counter = '%s:%s' % (release, package)
        counts = [random.randint(1,20) for r in xrange(6)]
        today = datetime.date.today()
        yesterday = today - datetime.timedelta(1)
        first_day = today - datetime.timedelta(6)
        dates = [x for x in
                    self.cassandra._date_range_iterator(first_day.strftime('%Y%m%d'),
                    yesterday.strftime('%Y%m%d'))]
        for pos, date in enumerate(dates):
            counters_cf.insert(counter, {date: counts[pos]})
        counters_cf.insert(counter, {today.strftime('%Y%m%d'): 25})
        crash_rate = self.cassandra.get_package_crash_rate(release, package,
                                                           'https://errors.ubuntu.com')
        self.assertEqual(crash_rate['increase'], True)

    def test_no_increase(self):
        '''If a package has less crashes today than the average there is not
           an increase.'''
        counters_cf = pycassa.ColumnFamily(self.pool, 'Counters')
        release = 'Ubuntu 12.10'
        package = 'whoopsie'
        counter = '%s:%s' % (release, package)
        counts = [random.randint(10,20) for r in xrange(6)]
        today = datetime.date.today()
        yesterday = today - datetime.timedelta(1)
        first_day = today - datetime.timedelta(6)
        dates = [x for x in
                    self.cassandra._date_range_iterator(first_day.strftime('%Y%m%d'),
                    yesterday.strftime('%Y%m%d'))]
        for pos, date in enumerate(dates):
            counters_cf.insert(counter, {date: counts[pos]})
        counters_cf.insert(counter, {today.strftime('%Y%m%d'): 5})
        crash_rate = self.cassandra.get_package_crash_rate(release, package,
                                                           'https://errors.ubuntu.com')
        self.assertEqual(crash_rate['increase'], False)

    def test_no_crashes(self):
        '''If a package has no crashes there is not an increase.'''
        release = 'Ubuntu 12.10'
        package = '2vcard'
        crash_rate = self.cassandra.get_package_crash_rate(release, package,
                                            'https://errors.ubuntu.com')
        self.assertEqual(crash_rate['increase'], False)

if __name__ == '__main__':
    unittest.main()
